  // EXPM()  Matrix exponential 
  // Paul Griffiths, August 21, 2009
  // Johns Hopkins University
  //
  // Matrix exponential algorithm from  Nicholas J. Higham, "THE SCALING AND SQUARING METHOD FOR THE
  // MATRIX EXPONENTIAL REVISITED", SIAM J. Matrix Anal. Appl., vol. 26, no. 4, pp. 1179 -- 1193.
  //
  // [X,d] = expm(A) 
  // [X,d] = expm(A,dmax), d is the degree of the Pade approximant
  // A:    Square matrix 
  // dmax: Maximum degree for the Pade approximant
  // X:    Computed matrix exponential of A
  // d:    Degree of the Pade approximant used. Should be 3, 5, 7, 9, or 13.
  // 
  // Performnace: .0013552 s for rand(20,20) matrix (12% slower than the builtin)
  // Builtin expm: .0012032 s 
  // Machine: 2.5 Ghz Intel Core 2 Duo CPU, Windows XP

function [X, d] = expm2(A, dmax)
  
  // Coefficients of degree-13 Pade approximant
  b = [64764752532480000, 32382376266240000, 7771770303897600, ...
       1187353796428800, 129060195264000, 10559470521600, ...
       670442572800, 33522128640, 1323241920, ...
       40840800, 960960, 16380, 182, 1];
       
  // Theshold values of ||A|| (1-norm) for the degrees of the Pade approximant
  // given in mvec.
  thetam = [ 1.495585217958292e-2, 2.539398330063230e-1,  ...
        9.504178996162932e-1, 2.097847961257068e0, 5.371920351148152e0 ];

  // Degrees of the Pade approximant.
  mvec = [ 3 5 7 9 13 ];
  
  // If dmax is provided, truncate mvec.
  if( argn(2) > 1 ) 
    if( dmax <= 13 & dmax >= 3 )
      idx = find( mvec <= dmax )
      mvec = mvec(idx);
      thetam = thetam(idx);
    else
      warning('dmax should be 3, 5, 7, 9, or 13.');
    end
  end
  
  
       
  // Check that A is square
  if( size(A,1) ~= size(A,2) )
    error("Usage : X = expm(A). A must be square");
  end
  n = size(A,1);
  
  // Preprocess x to reduce norm -- store mu and T for post-processing
  mu = trace(A)/n;
  A = A - mu*eye(n,n);
  [A, T] = balanc( A );
  
  // Select the appropriate degree for Pade approximant to expm().
  normA = norm(A,1); // store normA for later 
  idx = find( normA <= thetam );
  
  if( ~isempty(idx) )
      d = mvec(idx(1));
      s = 0; // no scaling
  else
    // Use the greatest Pade approximant (typically 13)
    //  and use scaling and squaring instead.
    d = mvec($);
    // Scale 
    s = ceil( log2( normA / thetam($) ) );    
    A = A/(2^s);
  end
  
  
  // Progressively compute up to a degree 13 approximant while minimizing the 
  //  number of matrix multiplications.
  
  // Full [13,13] approximant is given by:
  //   U = A*(A6*(b(14)*A6 + b(12)*A4 + b(10)*A2) + b(8)*A6 + b(6)*A4 + b(4)*A2 + b(2)*eye(n,n));
  //   V = A6*(b(13)*A6 + b(11)*A4 + b(9)*A2) + b(7)*A6 + b(5)*A4 + b(3)*A2 + b(1)*eye(n,n);
  //   X = inv(V-U)*(U+V);
  
  A2 = A*A;  
  U = b(4)*A2 + b(2)*eye(n,n);
  V = b(3)*A2 + b(1)*eye(n,n);
  if( d >= 5 )
    A4 = A2*A2;  
    U = U + b(6)*A4;
    V = V + b(5)*A4;
  end
  
  if( d >= 7 )    
    A6 = A2*A4;
    U = U + b(8)*A6;
    V = V + b(7)*A6;
  end

  if( d >= 9 ) 
    // compute terms to be multiplied by A6
    U6 = b(10)*A2;
    V6 = b(9)*A2;
    
    if( d == 13 ) // highest order
      U6 = U6 + b(14)*A6 + b(12)*A4;
      V6 = V6 + b(13)*A6 + b(11)*A4;
    end
    U6 = A6*U6;
    V6 = A6*V6;
  end
  
  U = A*U; // One more multiplication by A on all odd factors.
  
  X = inv(V-U)*(U+V); // Pade approximant.
  
  // Square to undo scaling
  for idx = 1:s
    X = X*X;
  end

  // Undo preprocessing
  X = exp(mu)*T*X*inv(T);
  
endfunction
